/*
 * CALLBACKS.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements functions to collect, display, sort, and deserialize
 * individual data types associated with files that we can enumerate.
 *
 * Copyright (c) 2014-15 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Sorting support
//

DWORD
SdirCompareAllocationSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->AllocationSize, (PULARGE_INTEGER)&Right->AllocationSize);
}

DWORD
SdirCompareArch (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->Architecture < Right->Architecture) {
        return SDIR_LESS_THAN;
    } else if (Left->Architecture > Right->Architecture) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareCompressionAlgorithm (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->CompressionAlgorithm < Right->CompressionAlgorithm) {
        return SDIR_LESS_THAN;
    } else if (Left->CompressionAlgorithm > Right->CompressionAlgorithm) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareCompressedFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->CompressedFileSize, (PULARGE_INTEGER)&Right->CompressedFileSize);
}

DWORD
SdirCompareCreateDate (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareDate(&Left->CreateTime, &Right->CreateTime);
}

DWORD
SdirCompareCreateTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->CreateTime, &Right->CreateTime);
}

DWORD
SdirCompareFileAttributes (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->FileAttributes < Right->FileAttributes) {
        return SDIR_LESS_THAN;
    } else if (Left->FileAttributes > Right->FileAttributes) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareFileExtension (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->Extension, Right->Extension);
}

DWORD
SdirCompareFileId (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileId, (PULARGE_INTEGER)&Right->FileId);
}

DWORD
SdirCompareFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->FileName, Right->FileName);
}

DWORD
SdirCompareFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileSize, (PULARGE_INTEGER)&Right->FileSize);
}

DWORD
SdirCompareFragmentCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FragmentCount, (PULARGE_INTEGER)&Right->FragmentCount);
}

DWORD
SdirCompareLinkCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->LinkCount < Right->LinkCount) {
        return SDIR_LESS_THAN;
    } else if (Left->LinkCount > Right->LinkCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareOsVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->OsVersionHigh < Right->OsVersionHigh) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionHigh > Right->OsVersionHigh) {
        return SDIR_GREATER_THAN;
    }

    if (Left->OsVersionLow < Right->OsVersionLow) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionLow > Right->OsVersionLow) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareReparseTag (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->ReparseTag < Right->ReparseTag) {
        return SDIR_LESS_THAN;
    } else if (Left->ReparseTag > Right->ReparseTag) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareShortName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->ShortFileName, Right->ShortFileName);
}

DWORD
SdirCompareStreamCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->StreamCount < Right->StreamCount) {
        return SDIR_LESS_THAN;
    } else if (Left->StreamCount > Right->StreamCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareUsn (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->Usn, (PULARGE_INTEGER)&Right->Usn);
}

DWORD
SdirCompareVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileVersion, (PULARGE_INTEGER)&Right->FileVersion);
}

DWORD
SdirCompareWriteDate (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareDate(&Left->WriteTime, &Right->WriteTime);
}

DWORD
SdirCompareWriteTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->WriteTime, &Right->WriteTime);
}

DWORD
SdirBitwiseFileAttributes (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if ((Left->FileAttributes & Right->FileAttributes) == Right->FileAttributes) {
        return SDIR_EQUAL;
    }
    return SDIR_NOT_EQUAL;
}

TCHAR
SdirGetUpcasedCharFromString (
    __in LPTSTR Str,
    __in DWORD Index
    )
{
    TCHAR Char = Str[Index];

    if (Char >= 'a' && Char <= 'z') {
        Char = Char - 'a' + 'A';
    }

    return Char;
}

DWORD
SdirBitwiseFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    DWORD LeftIndex, RightIndex;

    TCHAR CompareLeft;
    TCHAR CompareRight;

    LeftIndex = 0;
    RightIndex = 0;

    while (Left->FileName[LeftIndex] != '\0' && Right->FileName[RightIndex] != '\0') {

        CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);
        CompareRight = SdirGetUpcasedCharFromString(Right->FileName, RightIndex);

        LeftIndex++;
        RightIndex++;

        if (CompareRight == '?') {

            //
            //  '?' matches with everything.  We've already advanced to the next
            //  char, so continue.
            //

        } else if (CompareRight == '*') {

            //
            //  Skip one char so Right is the one after *.  Left should compare
            //  the character it's currently on.  Keep going through Left until
            //  we find the char in Right
            //

            LeftIndex--;
            CompareRight = SdirGetUpcasedCharFromString(Right->FileName, RightIndex);
            CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);

            while (CompareLeft != CompareRight && CompareLeft != '\0') {
                LeftIndex++;
                CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);
            }

        } else {
            if (CompareLeft != CompareRight) {
                return SDIR_NOT_EQUAL;
            }
        }

    }

    if (Left->FileName[LeftIndex] == '\0' && Right->FileName[RightIndex] == '\0') {
        return SDIR_EQUAL;
    }

    return SDIR_NOT_EQUAL;
}


//
//  File enumeration support
//

BOOL
SdirCollectAllocationSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    BOOL RealAllocSize = FALSE;

    if (Opts->GetFileInformationByHandleEx) {

        HANDLE hFile;

        hFile = CreateFile(FullPath,
                           FILE_READ_ATTRIBUTES,
                           FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                           NULL);
    
        if (hFile != INVALID_HANDLE_VALUE) {
            FILE_STANDARD_INFO StandardInfo;
    
            if (Opts->GetFileInformationByHandleEx(hFile, FileStandardInfo, &StandardInfo, sizeof(StandardInfo))) {
                Entry->AllocationSize = StandardInfo.AllocationSize;
                RealAllocSize = TRUE;
            }

            CloseHandle(hFile);
        }
    }

    if (!RealAllocSize) {
        ULONG BytesPerSector;
        ULONG SectorsPerCluster;
        ULONG FreeClusters;
        ULONG TotalClusters;
        ULONG ClusterSize;

        GetDiskFreeSpace(Opts->ParentName, &SectorsPerCluster, &BytesPerSector, &FreeClusters, &TotalClusters);

        ClusterSize = SectorsPerCluster * BytesPerSector;

        Entry->AllocationSize.LowPart = FindData->nFileSizeLow;
        Entry->AllocationSize.HighPart = FindData->nFileSizeHigh;

        SdirFileSizeFromLargeInt(&Entry->AllocationSize) = (SdirFileSizeFromLargeInt(&Entry->AllocationSize) + ClusterSize - 1) & (~(ClusterSize - 1));
    }

    return TRUE;
}

typedef struct _SDIR_PE_HEADERS {
    DWORD Signature;
    IMAGE_FILE_HEADER ImageHeader;
    IMAGE_OPTIONAL_HEADER OptionalHeader;
} SDIR_PE_HEADERS, *PSDIR_PE_HEADERS;

BOOL
SdirCapturePeHeaders (
    __in LPTSTR FullPath,
    __out PSDIR_PE_HEADERS PeHeaders
    )
{
    HANDLE hFileRead;
    IMAGE_DOS_HEADER DosHeader;
    DWORD BytesReturned;

    //
    //  We want the earlier handle to be attribute only so we can
    //  operate on directories, but we need data for this, so we
    //  end up with two handles.
    //

    hFileRead = CreateFile(FullPath,
                           FILE_READ_ATTRIBUTES|FILE_READ_DATA,
                           FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_BACKUP_SEMANTICS,
                           NULL);

    if (hFileRead != INVALID_HANDLE_VALUE) {

        if (ReadFile(hFileRead, &DosHeader, sizeof(DosHeader), &BytesReturned, NULL) &&
            BytesReturned == sizeof(DosHeader) &&
            DosHeader.e_magic == IMAGE_DOS_SIGNATURE &&
            DosHeader.e_lfanew != 0) {

            SetFilePointer(hFileRead, DosHeader.e_lfanew, NULL, FILE_BEGIN);

            if (ReadFile(hFileRead, PeHeaders, sizeof(SDIR_PE_HEADERS), &BytesReturned, NULL) &&
                BytesReturned == sizeof(SDIR_PE_HEADERS) &&
                PeHeaders->Signature == IMAGE_NT_SIGNATURE &&
                PeHeaders->ImageHeader.SizeOfOptionalHeader >= FIELD_OFFSET(IMAGE_OPTIONAL_HEADER, Subsystem)) {

                CloseHandle(hFileRead);
                return TRUE;

            }
        }
        CloseHandle(hFileRead);
    }
    return FALSE;
}

BOOL
SdirCollectArch (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    SDIR_PE_HEADERS PeHeaders;

    Entry->OsVersionHigh = 0;
    Entry->OsVersionLow = 0;

    if (SdirCapturePeHeaders(FullPath, &PeHeaders)) {

        Entry->Architecture = PeHeaders.ImageHeader.Machine;
    }

    return TRUE;
}

BOOL
SdirCollectCompressionAlgorithm (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->CompressionAlgorithm = SdirCompressionNone;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {

        USHORT NtfsCompressionAlgorithm;
        DWORD BytesReturned;
        struct {
            WOF_EXTERNAL_INFO WofHeader;
            union {
                WIM_PROVIDER_EXTERNAL_INFO WimInfo;
                FILE_PROVIDER_EXTERNAL_INFO FileInfo;
            } u;
        } WofInfo;

        if (DeviceIoControl(hFile, FSCTL_GET_COMPRESSION, NULL, 0, &NtfsCompressionAlgorithm, sizeof(NtfsCompressionAlgorithm), &BytesReturned, NULL)) {

            if (NtfsCompressionAlgorithm == COMPRESSION_FORMAT_LZNT1) {
                Entry->CompressionAlgorithm = SdirCompressionLznt;
            } else if (NtfsCompressionAlgorithm != COMPRESSION_FORMAT_NONE) {
                Entry->CompressionAlgorithm = SdirCompressionNtfsUnknown;
            }
        }

        if (Entry->CompressionAlgorithm == SdirCompressionNone) {
            if (DeviceIoControl(hFile, FSCTL_GET_EXTERNAL_BACKING, NULL, 0, &WofInfo, sizeof(WofInfo), &BytesReturned, NULL)) {
    
                if (WofInfo.WofHeader.Provider == WOF_PROVIDER_WIM) {
                    Entry->CompressionAlgorithm = SdirCompressionWim;
                } else if (WofInfo.WofHeader.Provider == WOF_PROVIDER_FILE) {
                    if (WofInfo.u.FileInfo.Algorithm == FILE_PROVIDER_COMPRESSION_XPRESS4K) {
                        Entry->CompressionAlgorithm = SdirCompressionXpress4k;
                    } else if (WofInfo.u.FileInfo.Algorithm == FILE_PROVIDER_COMPRESSION_XPRESS8K) {
                        Entry->CompressionAlgorithm = SdirCompressionXpress8k;
                    } else if (WofInfo.u.FileInfo.Algorithm == FILE_PROVIDER_COMPRESSION_XPRESS16K) {
                        Entry->CompressionAlgorithm = SdirCompressionXpress16k;
                    } else if (WofInfo.u.FileInfo.Algorithm == FILE_PROVIDER_COMPRESSION_LZX) {
                        Entry->CompressionAlgorithm = SdirCompressionLzx;
                    } else {
                        Entry->CompressionAlgorithm = SdirCompressionWofFileUnknown;
                    }
                } else {
                    Entry->CompressionAlgorithm = SdirCompressionWofUnknown;
                }
            }
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectCompressedFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->CompressedFileSize.LowPart = FindData->nFileSizeLow;
    Entry->CompressedFileSize.HighPart = FindData->nFileSizeHigh;

    if (Opts->GetCompressedFileSize) {
        Entry->CompressedFileSize.LowPart = Opts->GetCompressedFileSize(FullPath, &Entry->CompressedFileSize.HighPart);

        if (Entry->CompressedFileSize.LowPart == INVALID_FILE_SIZE) {
            Entry->CompressedFileSize.LowPart = FindData->nFileSizeLow;
            Entry->CompressedFileSize.HighPart = FindData->nFileSizeHigh;
        }
    }

    return TRUE;
}

BOOL
SdirCollectCreateTime (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftCreationTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->CreateTime);
    return TRUE;
}

BOOL
SdirCollectFileAttributes (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    DWORD i;

    //
    //  We do this bit by bit to ensure that we don't have file attributes
    //  recorded that we don't understand.  This allows us to perform
    //  equality comparisons where the result is understandable to the user
    //  in that it can be specified and displayed.
    //

    Entry->FileAttributes = 0;
    for (i = 0; i < SdirGetNumAttrPairs(); i++) {
        if (FindData->dwFileAttributes & AttrPairs[i].FileAttribute) {
            Entry->FileAttributes |= AttrPairs[i].FileAttribute;
        }
    }
    return TRUE;
}

BOOL
SdirCollectFileExtension (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    return TRUE;
}

BOOL
SdirCollectFileId (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->FileId.QuadPart = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->FileId.LowPart = FileInfo.nFileIndexLow;
            Entry->FileId.HighPart = FileInfo.nFileIndexHigh;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectFileName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileNameLengthInChars = (DWORD)_tcslen(FindData->cFileName);
    if (Entry->FileNameLengthInChars >= MAX_PATH - 1) {
        Entry->FileNameLengthInChars = MAX_PATH - 2;
    }
    if (Entry->FileNameLengthInChars > SdirDirCollectionLongest) {
        SdirDirCollectionLongest = Entry->FileNameLengthInChars;
    }

    SdirDirCollectionTotalNameLength += Entry->FileNameLengthInChars;

    memcpy(Entry->FileName, FindData->cFileName, Entry->FileNameLengthInChars * sizeof(TCHAR));
    Entry->FileName[Entry->FileNameLengthInChars] = '\0';

    Entry->Extension = _tcsrchr(Entry->FileName, '.');

    //
    //  For simplicity's sake, if we have no extension set the field
    //  to the end of string, so we'll see a valid pointer of nothing.
    //

    if (Entry->Extension == NULL) {
        Entry->Extension = Entry->FileName + Entry->FileNameLengthInChars;
    } else {
        Entry->Extension++;
    }

    return TRUE;
}

BOOL
SdirCollectFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileSize.LowPart = FindData->nFileSizeLow;
    Entry->FileSize.HighPart = FindData->nFileSizeHigh;
    return TRUE;
}

BOOL
SdirCollectFragmentCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->FragmentCount.HighPart = 0;
    Entry->FragmentCount.LowPart = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {

        STARTING_VCN_INPUT_BUFFER StartBuffer;
        union {
            RETRIEVAL_POINTERS_BUFFER Extents;
            UCHAR Buffer[2048];
        } u;
        DWORD BytesReturned;

        StartBuffer.StartingVcn.QuadPart = 0;

        while ((DeviceIoControl(hFile, FSCTL_GET_RETRIEVAL_POINTERS, &StartBuffer, sizeof(StartBuffer), &u.Extents, sizeof(u), &BytesReturned, NULL) || GetLastError() == ERROR_MORE_DATA) &&
               u.Extents.ExtentCount > 0) {

            StartBuffer.StartingVcn.QuadPart = u.Extents.Extents[u.Extents.ExtentCount - 1].NextVcn.QuadPart;

            //
            //  Check for overflow.  If so, bump highpart and continue.  We know we
            //  can't fit 4 billion extents in our stack buffer, so +1 works fine.
            //

            if (Entry->FragmentCount.LowPart + u.Extents.ExtentCount < Entry->FragmentCount.LowPart) {
                Entry->FragmentCount.HighPart++;
            }
            Entry->FragmentCount.LowPart += u.Extents.ExtentCount;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectLinkCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->LinkCount = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->LinkCount = FileInfo.nNumberOfLinks;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectOsVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    SDIR_PE_HEADERS PeHeaders;

    Entry->OsVersionHigh = 0;
    Entry->OsVersionLow = 0;

    if (SdirCapturePeHeaders(FullPath, &PeHeaders)) {

        Entry->OsVersionHigh = PeHeaders.OptionalHeader.MajorSubsystemVersion;
        Entry->OsVersionLow = PeHeaders.OptionalHeader.MinorSubsystemVersion;
    }

    return TRUE;
}

BOOL
SdirCollectReparseTag (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
        Entry->ReparseTag = FindData->dwReserved0;
    } else {
        Entry->ReparseTag = 0;
    }
    return TRUE;
}

BOOL
SdirCollectShortName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (FindData->cAlternateFileName[0] == '\0' && _tcslen(FindData->cFileName) <= 12) {
        CopyMemory(Entry->ShortFileName, FindData->cFileName, sizeof(FindData->cAlternateFileName));
    } else {
        CopyMemory(Entry->ShortFileName, FindData->cAlternateFileName, sizeof(FindData->cAlternateFileName));
    }
    return TRUE;
}

#ifdef UNICODE
BOOL
SdirCollectStreamCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFind;
    SDIR_WIN32_FIND_STREAM_DATA FindStreamData;

    Entry->StreamCount = 0;

    //
    //  These APIs are Unicode only.  We could do an ANSI to Unicode thunk here, but
    //  since Unicode is the default build and ANSI is only useful for older systems
    //  (where this API won't exist) there doesn't seem much point.
    //

    if (Opts->FindFirstStreamW && Opts->FindNextStreamW) {
        hFind = Opts->FindFirstStreamW(FullPath, 0, &FindStreamData, 0);
        if (hFind != INVALID_HANDLE_VALUE) {

            do {
                Entry->StreamCount++;
            } while (Opts->FindNextStreamW(hFind, &FindStreamData));
        }

        FindClose(hFind);
    }

    return TRUE;
}
#endif

BOOL
SdirCollectUsn (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->Usn.QuadPart = 0;
    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_OPEN_NO_RECALL,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {

        struct {
            SDIR_USN_RECORD UsnRecord;
            WCHAR FileName[MAX_PATH];
        } s1;
        DWORD BytesReturned;

        if (DeviceIoControl(hFile, FSCTL_READ_FILE_USN_DATA, NULL, 0, &s1, sizeof(s1), &BytesReturned, NULL)) {
            Entry->Usn.QuadPart = s1.UsnRecord.Usn.QuadPart;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    DWORD Junk;
    PVOID Buffer;
    DWORD VerSize = GetFileVersionInfoSize(FullPath, &Junk);
    VS_FIXEDFILEINFO * RootBlock;

    Entry->FileVersion.QuadPart = 0;
    Entry->FileVersionFlags = 0;

    Buffer = HeapAlloc(GetProcessHeap(), 0, VerSize);
    if (Buffer != NULL) {
        if (GetFileVersionInfo(FullPath, 0, VerSize, Buffer)) {
            if (VerQueryValue(Buffer, _T("\\"), &RootBlock, &Junk)) {
                Entry->FileVersion.HighPart = RootBlock->dwFileVersionMS;
                Entry->FileVersion.LowPart = RootBlock->dwFileVersionLS;
                Entry->FileVersionFlags = RootBlock->dwFileFlags & RootBlock->dwFileFlagsMask;
            }
        }
        HeapFree(GetProcessHeap(), 0, Buffer);
    }
    return TRUE;
}

BOOL
SdirCollectWriteTime(
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftLastWriteTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->WriteTime);
    return TRUE;
}

BOOL
SdirCollectSummary(
    __in PSDIR_DIRENT Entry
    )
{

    //
    //  Don't count . and .. at all
    //

    if ((Entry->FileNameLengthInChars == 1 && Entry->FileName[0] == '.') ||
        (Entry->FileNameLengthInChars == 2 && Entry->FileName[0] == '.' && Entry->FileName[1] == '.')) {

        return TRUE;
    }

    //
    //  Otherwise, count either as a file or a dir
    //

    if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
        Summary->NumDirs++;
    } else {
        Summary->NumFiles++;
    }

    Summary->TotalSize += SdirFileSizeFromLargeInt(&Entry->FileSize);

    if (Opts->FtCompressedFileSize.Flags & SDIR_FEATURE_COLLECT) {
        Summary->CompressedSize += SdirFileSizeFromLargeInt(&Entry->CompressedFileSize);
    }

    return TRUE;
}

//
//  When criteria are specified to apply attributes, we need to load the specification
//  into a dummy dirent to perform comparisons against.  The below functions implement
//  these.
//

BOOL
SdirGenerateAllocationSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->AllocationSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateArch(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->Architecture = 0;

    if (_tcsicmp(String, _T("i386")) == 0) {
        Entry->Architecture = IMAGE_FILE_MACHINE_I386;
    } else if (_tcsicmp(String, _T("amd64")) == 0) {
        Entry->Architecture = IMAGE_FILE_MACHINE_AMD64;
    } else if (_tcsicmp(String, _T("arm")) == 0) {
        Entry->Architecture = IMAGE_FILE_MACHINE_ARMNT;
    } else if (_tcsicmp(String, _T("arm64")) == 0) {
        Entry->Architecture = IMAGE_FILE_MACHINE_ARM64;
    }
    return TRUE;
}

BOOL
SdirGenerateCompressionAlgorithm(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->CompressionAlgorithm = SdirCompressionNone;

    if (_tcsicmp(String, _T("LZNT")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionLznt;
    } else if (_tcsicmp(String, _T("NTFS")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionNtfsUnknown;
    } else if (_tcsicmp(String, _T("WIM")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionWim;
    } else if (_tcsicmp(String, _T("LZX")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionLzx;
    } else if (_tcsicmp(String, _T("Xp4")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionXpress4k;
    } else if (_tcsicmp(String, _T("Xp8")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionXpress8k;
    } else if (_tcsicmp(String, _T("Xp16")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionXpress16k;
    } else if (_tcsicmp(String, _T("File")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionWofFileUnknown;
    } else if (_tcsicmp(String, _T("Wof")) == 0) {
        Entry->CompressionAlgorithm = SdirCompressionWofUnknown;
    }
    return TRUE;
}

BOOL
SdirGenerateCompressedFileSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->CompressedFileSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateCreateDate(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToDate(String, &Entry->CreateTime);
}

BOOL
SdirGenerateCreateTime(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToTime(String, &Entry->CreateTime);
}

BOOL
SdirGenerateFileAttributes(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    DWORD i;

    Entry->FileAttributes = 0;

    while (*String) {

        for (i = 0; i < SdirGetNumAttrPairs(); i++) {
            if (*String == AttrPairs[i].DisplayLetter) {
                Entry->FileAttributes |= AttrPairs[i].FileAttribute;
            }
        }

        String++;
    }
    return TRUE;
}

BOOL
SdirGenerateFileExtension(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    //
    //  Since we have one dirent per comparison, just shove the extension in
    //  the file name buffer and point the extension to it.  This buffer can't
    //  be used for anything else anyway.
    //

    _stprintf_s(Entry->FileName, sizeof(Entry->FileName)/sizeof(Entry->FileName[0]), _T("%s"), String);
    Entry->FileName[sizeof(Entry->FileName)/sizeof(Entry->FileName[0]) - 1] = '\0';
    Entry->Extension = Entry->FileName;

    return TRUE;
}

BOOL
SdirGenerateFileName(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->FileNameLengthInChars = _stprintf_s(Entry->FileName, sizeof(Entry->FileName)/sizeof(Entry->FileName[0]), _T("%s"), String);
    Entry->FileName[sizeof(Entry->FileName)/sizeof(Entry->FileName[0]) - 1] = '\0';

    return TRUE;
}

BOOL
SdirGenerateFileSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->FileSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateFragmentCount(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->FragmentCount = SdirStringToNum64(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateLinkCount(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->LinkCount = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateOsVersion(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    TCHAR * endptr = NULL;

    Entry->OsVersionHigh = (WORD)SdirStringToNum32(String, &endptr);

    if (*endptr == '.') {
        Entry->OsVersionLow = (WORD)SdirStringToNum32(endptr+1, NULL);
    }

    return TRUE;
}

BOOL
SdirGenerateReparseTag(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->ReparseTag = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateShortName(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    _stprintf_s(Entry->ShortFileName, sizeof(Entry->ShortFileName)/sizeof(Entry->ShortFileName[0]), _T("%s"), String);
    Entry->ShortFileName[sizeof(Entry->ShortFileName)/sizeof(Entry->ShortFileName[0]) - 1] = '\0';

    return TRUE;
}

BOOL
SdirGenerateStreamCount(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->StreamCount = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateUsn(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->Usn = SdirStringToNum64(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateVersion(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    LARGE_INTEGER Version = {0};
    TCHAR * endptr = NULL;

    Version.HighPart = SdirStringToNum32(String, &endptr)<<16;

    if (*endptr == '.') {
        Version.HighPart = Version.HighPart + (WORD)SdirStringToNum32(endptr+1, &endptr);

        if (*endptr == '.') {
            Version.LowPart = SdirStringToNum32(endptr+1, &endptr)<<16;
            if (*endptr == '.') {
                Version.LowPart = Version.LowPart + (WORD)SdirStringToNum32(endptr+1, NULL);
            }
                
        }
    }

    Entry->FileVersion = Version;

    return TRUE;
}

BOOL
SdirGenerateWriteDate(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToDate(String, &Entry->WriteTime);
}

BOOL
SdirGenerateWriteTime(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToTime(String, &Entry->WriteTime);
}


//
//  Specific formatting and sizing callbacks for each supported piece of file metadata.
//

ULONG
SdirDisplayAllocationSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayGenericSize(Buffer, Attributes, &Entry->AllocationSize);
}

ULONG
SdirDisplayArch (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR StrBuf[8];
    LPTSTR StrFixed;

    if (Buffer) {
        switch(Entry->Architecture) {
            case IMAGE_FILE_MACHINE_UNKNOWN:
                StrFixed = _T("None ");
                break;
            case IMAGE_FILE_MACHINE_I386:
                StrFixed = _T("i386 ");
                break;
            case IMAGE_FILE_MACHINE_AMD64:
                StrFixed = _T("amd64");
                break;
            case IMAGE_FILE_MACHINE_ARMNT:
                StrFixed = _T("arm  ");
                break;
            case IMAGE_FILE_MACHINE_ARM64:
                StrFixed = _T("arm64");
                break;
            default:
                StrFixed = _T("New? ");
                break;
        }
        _stprintf_s(StrBuf, sizeof(StrBuf)/sizeof(StrBuf[0]), _T(" %5s"), StrFixed);
        SdirPasteStr(Buffer, StrBuf, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayCompressionAlgorithm (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR StrBuf[8];
    LPTSTR StrFixed;

    if (Buffer) {
        switch(Entry->CompressionAlgorithm) {
            case SdirCompressionNone:
                StrFixed = _T("None");
                break;
            case SdirCompressionLznt:
                StrFixed = _T("LZNT");
                break;
            case SdirCompressionNtfsUnknown:
                StrFixed = _T("NTFS");
                break;
            case SdirCompressionWim:
                StrFixed = _T("WIM ");
                break;
            case SdirCompressionLzx:
                StrFixed = _T("LZX ");
                break;
            case SdirCompressionXpress4k:
                StrFixed = _T("Xp4 ");
                break;
            case SdirCompressionXpress8k:
                StrFixed = _T("Xp8 ");
                break;
            case SdirCompressionXpress16k:
                StrFixed = _T("Xp16");
                break;
            case SdirCompressionWofFileUnknown:
                StrFixed = _T("File");
                break;
            case SdirCompressionWofUnknown:
                StrFixed = _T("Wof ");
                break;
            default:
                StrFixed = _T("BUG ");
                break;
        }
        _stprintf_s(StrBuf, sizeof(StrBuf)/sizeof(StrBuf[0]), _T(" %4s"), StrFixed);
        SdirPasteStr(Buffer, StrBuf, Attributes, 5);
    }
    return 5;
}

ULONG
SdirDisplayCompressedFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayGenericSize(Buffer, Attributes, &Entry->CompressedFileSize);
}

ULONG
SdirDisplayFileCreateDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, Attributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileCreateTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, Attributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileAttributes (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR StrAtts[32];
    DWORD i;

    if (Buffer) {
        StrAtts[0] = ' ';
        for (i = 0; i < SdirGetNumAttrPairs(); i++) {
            if (Entry->FileAttributes & AttrPairs[i].FileAttribute) {
                StrAtts[i + 1] = AttrPairs[i].DisplayLetter;
            } else {
                StrAtts[i + 1] = '-';
            }
        }

        StrAtts[i + 1] = '\0';
    
        SdirPasteStr(Buffer, StrAtts, Attributes, SdirGetNumAttrPairs() + 1);
    }
    return SdirGetNumAttrPairs() + 1;
}

ULONG
SdirDisplayFileId (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, Attributes, &Entry->FileId);
}

ULONG
SdirDisplayFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    if (Buffer) {
        if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
            if (Entry->ReparseTag == IO_REPARSE_TAG_SYMLINK) {
                SdirPasteStr(Buffer, _T(" <LNK>"), Entry->RenderAttributes, 6);
            } else if (Entry->ReparseTag == IO_REPARSE_TAG_MOUNT_POINT) {
                SdirPasteStr(Buffer, _T(" <MNT>"), Entry->RenderAttributes, 6);
            } else {
                return SdirDisplayGenericSize(Buffer, Attributes, &Entry->FileSize);
            }
            return 6;
        } else if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            SdirPasteStr(Buffer, _T(" <DIR>"), Entry->RenderAttributes, 6);
            return 6;
        } else {
            return SdirDisplayGenericSize(Buffer, Attributes, &Entry->FileSize);
        }
    }
    return 6;
}

ULONG
SdirDisplayFragmentCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[8];

    if (Buffer) {

        //
        //  As a special hack to suppress the 'b' suffix, just display as
        //  a number unless it's six digits.  At that point we know we'll
        //  have a 'k' suffix or above.
        //

        if (SdirFileSizeFromLargeInt(&Entry->FragmentCount) <= 99999) { 
            _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %05i"), Entry->FragmentCount.LowPart);
            SdirPasteStr(Buffer, Str, Attributes, 6);
        } else {
            SdirDisplayGenericSize(Buffer, Attributes, &Entry->FragmentCount);
        }
    }
    return 6;
}

ULONG
SdirDisplayLinkCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        if (Entry->LinkCount >= 1000) {
            _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" >1k"));
        } else {
            _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->LinkCount);
        }
        SdirPasteStr(Buffer, Str, Attributes, 4);
    }
    return 4;
}

ULONG
SdirDisplayOsVersion (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[30];
    BYTE  ThisOsMajor = LOBYTE(LOWORD(Opts->OsVersion));
    BYTE  ThisOsMinor = HIBYTE(LOWORD(Opts->OsVersion));

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %02i.%02i"), Entry->OsVersionHigh, Entry->OsVersionLow);
    
        if (Entry->OsVersionHigh > ThisOsMajor ||
            Entry->OsVersionHigh == ThisOsMajor && Entry->OsVersionLow > ThisOsMinor) {
    
            Attributes = SDIR_FUTURE_VERSION_COLOR;
        }
    
        SdirPasteStr(Buffer, Str, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayReparseTag (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG Tag = 0;
    if (Entry) {
        Tag = Entry->ReparseTag;
    }
    return SdirDisplayHex32(Buffer, Attributes, Tag);
}

ULONG
SdirDisplayShortName (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG CurrentChar = 0;
    DWORD ShortLength;

    //
    //  Just because we like special cases, add a space only if both
    //  names are being displayed.
    //

    if (Opts->FtFileName.Flags & SDIR_FEATURE_DISPLAY) {
        if (Buffer) {
            SdirPasteStrAndPad(&Buffer[CurrentChar], NULL, 0, 0, 1);
        }
        CurrentChar++;
    }

    if (Buffer) {
        ShortLength = (DWORD)_tcslen(Entry->ShortFileName);

        SdirPasteStrAndPad(&Buffer[CurrentChar], Entry->ShortFileName, Attributes, ShortLength, 12);
    }
    CurrentChar += 12;
    return CurrentChar;
}

ULONG
SdirDisplayStreamCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        if (Entry->StreamCount >= 1000) {
            _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" >1k"));
        } else {
            _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->StreamCount);
        }
        SdirPasteStr(Buffer, Str, Attributes, 4);
    }
    return 4;
}

ULONG
SdirDisplayUsn(
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, Attributes, &Entry->Usn);
}

ULONG
SdirDisplayVersion(
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[40];

    if (Buffer) {
        _stprintf_s(Str,
                    sizeof(Str)/sizeof(Str[0]),
                    _T(" %02i.%02i.%05i.%05i %c%c"),
                    Entry->FileVersion.HighPart >> 16,
                    Entry->FileVersion.HighPart & 0xffff,
                    Entry->FileVersion.LowPart >> 16,
                    Entry->FileVersion.LowPart & 0xffff,
                    Entry->FileVersionFlags & VS_FF_DEBUG ? 'D' : '-',
                    Entry->FileVersionFlags & VS_FF_PRERELEASE ? 'P' : '-');

        SdirPasteStr(Buffer, Str, Attributes, 21);
    }
    return 21;
}

ULONG
SdirDisplaySummary(
    __in WORD DefaultAttributes
    )
{
    SDIR_FMTCHAR Buffer[200];
    TCHAR Str[100];
    DWORD Len;
    LARGE_INTEGER Size;
    ULONG CurrentChar = 0;

    _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %i "), Summary->NumFiles);
    Len = (DWORD)_tcslen(Str);
    SdirPasteStr(&Buffer[CurrentChar], Str, Opts->FtNumberFiles.HighlightColor, Len);
    CurrentChar += Len;

    SdirPasteStr(&Buffer[CurrentChar], _T("files,"), DefaultAttributes, 6);
    CurrentChar += 6;

    _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %i "), Summary->NumDirs);
    Len = (DWORD)_tcslen(Str);
    SdirPasteStr(&Buffer[CurrentChar], Str, Opts->FtNumberFiles.HighlightColor, Len);
    CurrentChar += Len;

    SdirPasteStr(&Buffer[CurrentChar], _T("dirs,"), DefaultAttributes, 5);
    CurrentChar += 5;

    Size.HighPart = 0;
    SdirFileSizeFromLargeInt(&Size) = Summary->TotalSize;

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], Opts->FtFileSize.HighlightColor, &Size);
    SdirPasteStr(&Buffer[CurrentChar], _T(" used,"), DefaultAttributes, 6);
    CurrentChar += 6;

    if (Opts->FtCompressedFileSize.Flags & SDIR_FEATURE_DISPLAY) {
        Size.HighPart = 0;
        SdirFileSizeFromLargeInt(&Size) = Summary->CompressedSize;

        CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], Opts->FtCompressedFileSize.HighlightColor, &Size);
        SdirPasteStr(&Buffer[CurrentChar], _T(" compressed,"), DefaultAttributes, 12);
        CurrentChar += 12;
    }

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], Opts->FtFileSize.HighlightColor, &Summary->VolumeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol size,"), DefaultAttributes, 10);
    CurrentChar += 10;

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], Opts->FtFileSize.HighlightColor, &Summary->FreeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol free"), DefaultAttributes, 9);
    CurrentChar += 9;

    SdirWrite(Buffer, CurrentChar);

    return CurrentChar;
}

ULONG
SdirDisplayFileWriteDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, Attributes, &Entry->WriteTime);
}

ULONG
SdirDisplayFileWriteTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, Attributes, &Entry->WriteTime);
}

//
//  The below table corresponds to the supported options in the program.
//  Each option has a flags value describing whether to collect or display
//  the piece of metadata, a default, a switch to turn display on or off
//  or sort by it, a callback function to tell how much space it will need,
//  an optional sort compare function, and some help text.
//
//  File names, extensions and attributes are always collected, because
//  that's how we determine colors.
//

#define OPT_OS(x) FIELD_OFFSET(SDIR_OPTS, x)

const SDIR_OPT SdirOptions[] = {
    {OPT_OS(FtAllocationSize),           _T("as"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        SdirDisplayAllocationSize,       SdirCollectAllocationSize,
        SdirCompareAllocationSize,       NULL,                          SdirGenerateAllocationSize,
        "allocation size"},

    {OPT_OS(FtArch),                     _T("ar"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        SdirDisplayArch,                 SdirCollectArch,
        SdirCompareArch,                 NULL,                          SdirGenerateArch,
        "CPU architecture"},

    {OPT_OS(FtBriefAlternate),           _T("ba"), {0, BACKGROUND_BLUE},
        NULL,                            NULL,                 
        NULL,                            NULL,                          NULL,
        "brief alternate mask"},

    {OPT_OS(FtCompressionAlgorithm),     _T("ca"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayCompressionAlgorithm, SdirCollectCompressionAlgorithm,
        SdirCompareCompressionAlgorithm, NULL,                          SdirGenerateCompressionAlgorithm,
        "compression algorithm"},
    
    {OPT_OS(FtCreateDate),               _T("cd"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFileCreateDate,       SdirCollectCreateTime,
        SdirCompareCreateDate,           NULL,                          SdirGenerateCreateDate,
        "create date"},

    {OPT_OS(FtCompressedFileSize),       _T("cs"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        SdirDisplayCompressedFileSize,   SdirCollectCompressedFileSize,
        SdirCompareCompressedFileSize,   NULL,                          SdirGenerateCompressedFileSize,
        "compressed size"},

    {OPT_OS(FtCreateTime),               _T("ct"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFileCreateTime,       SdirCollectCreateTime,
        SdirCompareCreateTime,           NULL,                          SdirGenerateCreateTime,
        "create time"},

    {OPT_OS(FtError),                     _T("er"), {0, FOREGROUND_RED|FOREGROUND_INTENSITY},
        NULL,                            NULL,                 
        NULL,                            NULL,                          NULL,
        "errors"},

    {OPT_OS(FtFileAttributes),           _T("fa"), {SDIR_FEATURE_COLLECT|SDIR_FEATURE_ALLOW_DISPLAY, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        SdirDisplayFileAttributes,       SdirCollectFileAttributes,
        SdirCompareFileAttributes,       SdirBitwiseFileAttributes,     SdirGenerateFileAttributes,
        "file attributes"},

    {OPT_OS(FtFragmentCount),            _T("fc"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFragmentCount,        SdirCollectFragmentCount,
        SdirCompareFragmentCount,        NULL,                          SdirGenerateFragmentCount,
        "fragment count"},

    {OPT_OS(FtFileExtension),            _T("fe"), {SDIR_FEATURE_DISPLAY|SDIR_FEATURE_COLLECT|SDIR_FEATURE_ALLOW_SORT|SDIR_FEATURE_FIXED_COLOR, 0},
        NULL,                            SdirCollectFileExtension,
        SdirCompareFileExtension,        NULL,                          SdirGenerateFileExtension,
        "file extension"},

    {OPT_OS(FtFileId),                   _T("fi"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFileId,               SdirCollectFileId,
        SdirCompareFileId,               NULL,                          NULL,
        "file id"},

    {OPT_OS(FtFileName),                 _T("fn"), {SDIR_FEATURE_DISPLAY|SDIR_FEATURE_COLLECT|SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT|SDIR_FEATURE_USE_FILE_COLOR, 0},
        NULL,                            SdirCollectFileName,
        SdirCompareFileName,             SdirBitwiseFileName,           SdirGenerateFileName,
        "file name"},

    {OPT_OS(FtFileSize),                 _T("fs"), {SDIR_FEATURE_DISPLAY|SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        SdirDisplayFileSize,             SdirCollectFileSize,
        SdirCompareFileSize,             NULL,                          SdirGenerateFileSize,
        "file size"},

    {OPT_OS(FtGrid),                     _T("gr"), {0, FOREGROUND_GREEN},
        NULL,                            NULL,                 
        NULL,                            NULL,                          NULL,
        "grid"},

    {OPT_OS(FtLinkCount),                _T("lc"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayLinkCount,            SdirCollectLinkCount,
        SdirCompareLinkCount,            NULL,                          SdirGenerateLinkCount,
        "link count"},

    {OPT_OS(FtNumberFiles),              _T("nf"), {0, FOREGROUND_GREEN|FOREGROUND_INTENSITY},
        NULL,                            NULL,                 
        NULL,                            NULL,                          NULL,
        "number files"},

#ifdef UNICODE
    {OPT_OS(FtNamedStreams),             _T("ns"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_FIXED_COLOR, 0},
        NULL,                            NULL,                   
        NULL,                            NULL,                          NULL,
        "named streams"},
#endif

    {OPT_OS(FtOsVersion),                _T("os"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayOsVersion,            SdirCollectOsVersion,
        SdirCompareOsVersion,            NULL,                          SdirGenerateOsVersion,
        "minimum OS version"},

    {OPT_OS(FtReparseTag),               _T("rt"), {SDIR_FEATURE_COLLECT|SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayReparseTag,           SdirCollectReparseTag,
        SdirCompareReparseTag,           NULL,                          SdirGenerateReparseTag,
        "reparse tag"},

#ifdef UNICODE
    {OPT_OS(FtStreamCount),              _T("sc"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayStreamCount,          SdirCollectStreamCount,
        SdirCompareStreamCount,          NULL,                          SdirGenerateStreamCount,
        "stream count"},
#endif

    {OPT_OS(FtSummary),                  _T("sm"), {SDIR_FEATURE_DISPLAY|SDIR_FEATURE_COLLECT|SDIR_FEATURE_ALLOW_DISPLAY, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE},
        NULL,                            NULL,
        NULL,                            NULL,                          NULL,
        "summary"},

    {OPT_OS(FtShortName),                _T("sn"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT|SDIR_FEATURE_USE_FILE_COLOR, 0},
        SdirDisplayShortName,            SdirCollectShortName,
        SdirCompareShortName,            NULL,                          SdirGenerateShortName,
        "short name"},

    {OPT_OS(FtUsn),                      _T("us"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayUsn,                  SdirCollectUsn,
        SdirCompareUsn,                  NULL,                          SdirGenerateUsn,
        "USN"},

    {OPT_OS(FtVersion),                  _T("vr"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayVersion,              SdirCollectVersion,
        SdirCompareVersion,              NULL,                          SdirGenerateVersion,
        "version"},

    {OPT_OS(FtWriteDate),                _T("wd"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFileWriteDate,        SdirCollectWriteTime,
        SdirCompareWriteDate,            NULL,                          SdirGenerateWriteDate,
        "write date"},

    {OPT_OS(FtWriteTime),                _T("wt"), {SDIR_FEATURE_ALLOW_DISPLAY|SDIR_FEATURE_ALLOW_SORT, FOREGROUND_RED|FOREGROUND_GREEN},
        SdirDisplayFileWriteTime,        SdirCollectWriteTime,
        SdirCompareWriteTime,            NULL,                          SdirGenerateWriteTime,
        "write time"},
};

DWORD
SdirGetNumSdirOptions()
{
    return sizeof(SdirOptions)/sizeof(SdirOptions[0]);
}

//
//  This table is really quite redundant, but it corresponds to the order
//  in which we will render each piece of metadata.  The table above is
//  the order we display help.
//


const SDIR_EXEC SdirExec[] = {
    {OPT_OS(FtFileId),               SdirDisplayFileId},
    {OPT_OS(FtUsn),                  SdirDisplayUsn},
    {OPT_OS(FtFileSize),             SdirDisplayFileSize},
    {OPT_OS(FtCompressedFileSize),   SdirDisplayCompressedFileSize},
    {OPT_OS(FtAllocationSize),       SdirDisplayAllocationSize},
    {OPT_OS(FtFileAttributes),       SdirDisplayFileAttributes},
    {OPT_OS(FtReparseTag),           SdirDisplayReparseTag},
    {OPT_OS(FtLinkCount),            SdirDisplayLinkCount},
    {OPT_OS(FtStreamCount),          SdirDisplayStreamCount},
    {OPT_OS(FtCreateDate),           SdirDisplayFileCreateDate},
    {OPT_OS(FtCreateTime),           SdirDisplayFileCreateTime},
    {OPT_OS(FtWriteDate),            SdirDisplayFileWriteDate},
    {OPT_OS(FtWriteTime),            SdirDisplayFileWriteTime},
    {OPT_OS(FtVersion),              SdirDisplayVersion},
    {OPT_OS(FtOsVersion),            SdirDisplayOsVersion},
    {OPT_OS(FtArch),                 SdirDisplayArch},
    {OPT_OS(FtCompressionAlgorithm), SdirDisplayCompressionAlgorithm},
    {OPT_OS(FtFragmentCount),        SdirDisplayFragmentCount},
};

DWORD
SdirGetNumSdirExec()
{
    return sizeof(SdirExec)/sizeof(SdirExec[0]);
}

// vim:sw=4:ts=4:et:
